/*
  hardware_avr.cpp - Hardware Library for AVR

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifdef __AVR__

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include "hardware.h"
#include "alarm.h"

#ifdef TIFR
#define TIFR0 TIFR
#endif

#ifdef TIMSK
#define TIMSK0 TIMSK
#endif

#if (F_CPU == 16000000) || (F_CPU == 16500000)
#define PERIOD_US 4
#elif (F_CPU == 8000000)
#define PERIOD_US 8
#else
#error "System Clock Not Supported. (8MHz or 16MHz)"
#endif

inline void AlarmClass_timeup(void)
{
  if (!AlarmTimer.timeup())
    TIMSK0 &= ~_BV(OCIE0B);
}

ISR(TIMER0_COMPA_vect)
{
  AlarmClass_timeup();
}

void Hardware::setupAlarmTimer(void)
{
  TCCR0A = 0; // standard mode
}

void Hardware::startAlarmTimer(int32_t interval)
{
  uint8_t save = disableAndSaveInterrupts();
  if (interval <= 2 * PERIOD_US)
    OCR0A = TCNT0 + 2;
  else if (interval <= 256 * PERIOD_US)
    OCR0A = TCNT0 + (uint8_t)((uint16_t)interval / PERIOD_US) - 1;
  else if (!(TIMSK0 & _BV(OCIE0A)))
    OCR0A = TCNT0 - 1;
  TIFR0 = _BV(OCF0A);
  TIMSK0 |= _BV(OCIE0A);
  restoreInterrupts(save);
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  uint8_t save = SREG;
  cli();
  return save;
}

void Hardware::restoreInterrupts(size_t save)
{
  SREG = save;
}

void Hardware::idle(void)
{
  sei();
  set_sleep_mode(SLEEP_MODE_IDLE);
  sleep_mode();
}

#endif
